/*
 * Copyright (C) 2023  Valentin Lorentz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License version 3,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

use enum_dispatch::enum_dispatch;
use ratatui::prelude::*;
use ratatui::widgets::Widget;

mod bottom_aligned_container;
pub use bottom_aligned_container::BottomAlignedContainer;
mod bottom_aligned_paragraph;
pub use bottom_aligned_paragraph::BottomAlignedParagraph;

pub(crate) mod prerender;
pub use prerender::Prerender;

mod divider;
pub use divider::Divider;

#[rustfmt::skip] // reflow is vendored from ratatui, let's avoid changes
mod reflow;

/// A [`Widget`] that returns how many columns and lines it needs to draw everything
/// (which is the number of lines it actually drew if it fits on screen)
#[enum_dispatch]
pub trait OverlappableWidget {
  /// Returns how many lines, from the bottom of the area, this widget would actually draw
  /// if rendered with the given width
  fn height(&self, width: u16) -> u64;

  /// Render the widget from the bottom, and return the width and height of the area
  /// actually drawn.
  fn render_overlap(self, area: Rect, buf: &mut Buffer) -> (u16, u16);
}

/// Enum of [`OverlappableWidget`] implementors
#[enum_dispatch(OverlappableWidget)]
pub enum BacklogItemWidget<'a> {
  Paragraph(BottomAlignedParagraph<'a>),
  Container(BottomAlignedContainer<'a>),
  Divider(Divider<'a>),
  #[cfg(feature = "images")]
  Image(BottomAlignedImage<'a>),
  Empty(EmptyWidget),
}

#[derive(Clone)]
pub struct EmptyWidget;

impl OverlappableWidget for EmptyWidget {
  fn height(&self, _width: u16) -> u64 {
    0
  }
  fn render_overlap(self, _area: Rect, _buf: &mut Buffer) -> (u16, u16) {
    (0, 0)
  }
}

#[cfg(feature = "images")]
mod images {
  use std::sync::Mutex;

  use ratatui_image::picker::Picker;

  use super::*;

  lazy_static::lazy_static! {
    static ref PROTOCOL_PICKER_LOCK: Mutex<()> = Mutex::new(());
    pub static ref PROTOCOL_PICKER: Picker = make_protocol_picker();
  }

  fn make_protocol_picker() -> Picker {
    let guard = PROTOCOL_PICKER_LOCK.lock().unwrap();
    let mut picker = ratatui_image::picker::Picker::from_termios()
      .expect("ratatui_image::picker::Picker::from_termios() failed");
    picker.guess_protocol();
    drop(guard);
    picker
  }

  pub struct BottomAlignedImage<'a>(&'a image::DynamicImage);

  impl<'a> BottomAlignedImage<'a> {
    pub fn new(image: &'a image::DynamicImage) -> Self {
      BottomAlignedImage(image)
    }
  }

  impl<'a> OverlappableWidget for BottomAlignedImage<'a> {
    fn height(&self, width: u16) -> u64 {
      let width_px: u32 = u32::min(
        self.0.width(),
        (width as u32) * (PROTOCOL_PICKER.font_size.0 as u32),
      );
      let height_px = u32::min(self.0.height(), width_px * self.0.height() / self.0.width());
      u32::min(
        u16::MAX.into(),
        height_px / (PROTOCOL_PICKER.font_size.1 as u32),
      )
      .into()
    }

    fn render_overlap(self, mut area: Rect, buf: &mut Buffer) -> (u16, u16) {
      let width = u16::min(
        u32::min(
          u16::MAX.into(),
          self.0.width() / (PROTOCOL_PICKER.font_size.0 as u32),
        ) as u16,
        area.width,
      );
      let height = u64::min(self.height(width), u16::MAX.into()) as u16;

      area.y = u16::max(
        area.y,
        area.y.saturating_add(area.height).saturating_sub(height),
      );

      let protocol = PROTOCOL_PICKER
        .clone()
        .new_protocol(
          self.0.clone(),
          Rect {
            x: area.x,
            y: area.y,
            width,
            height,
          },
          ratatui_image::Resize::Crop,
        )
        .expect("picker.new_protocol failed");
      let widget = ratatui_image::Image::new(protocol.as_ref());
      widget.render(area, buf); // TODO: scroll
      (area.width, 20)
    }
  }
}

#[cfg(feature = "images")]
pub use images::*;

/*
impl<W: OverlappableWidget> Widget for W {
  fn render(self, area: Rect, buf: &mut Buffer) {
    self.render_overlap(area, buf);
  }
}
*/
